Promise	= require 'bluebird'
Emitter	= require('events').EventEmitter
path	= require 'path'
cluster	= require 'cluster'
fs		= require 'fs'
os		= require 'os'
http	= require 'http'
require 'colors'

{ clone, merge, typeOf, toArray } = utils = require './helpers/utils'
merge utils, require './helpers/utils-node'

module.exports = class Lance
	constructor: (newCfg) ->
		@cfg = clone @cfg
		merge @cfg, newCfg if newCfg

		if @cfg.catchUncaught
			@cfg.Promise.onPossiblyUnhandledRejection @onPossiblyUnhandledRejection.bind this if @cfg.Promise
			Promise.onPossiblyUnhandledRejection @onPossiblyUnhandledRejection.bind this

		#
		# Emitter, events
		#

		@emitter = new @Emitter()
		@extendEmitter this

		@helpers.events this, this

		#
		# Format the config
		#

		if @cfg.server.cluster is null
			@cfg.server.cluster = ! @cfg.server.workers < 2

		@paths =
			root: @cfg.root or path.dirname require.main.filename

		#if @cfg.createTemplater
		templater = @cfg.templater or @cfg.templating

		if templater.autoConstruct
			templater.root	= @paths.root if not templater.root
			@templater		= new @Templater templater, this

		@cache		= new @Cache @cfg.cache
		@router		= new @Router @cfg.router, this

	onPossiblyUnhandledRejection: (err) ->
		@emit 'error', err, 'onPossiblyUnhandledRejection'

	createServer: (requestCallback) ->
		@requestCallback = requestCallback or ->

		http.globalAgent.maxSockets = @cfg.server.maxSockets or 20
		
		@server = http.createServer (req, res) => new @RequestHandler req, res, this

		if not cluster.isWorker
			@emit 'server.starting', this

		@server.on 'listening', =>
			if @cfg.server.method is 'socket' and @cfg.server.socketPerms
				fs.chmod @cfg.server.socket, @cfg.server.socketPerms

			if not cluster.isWorker or cluster.worker.id is 1
				@emit 'server.listening', this

		return @server

	createCluster: -> new Promise (resolve, reject) =>
		@cfg.server.method = 'port' if @cfg.server.method is 'socket' and not @cfg.server.socket

		if cluster.isWorker
			server = @createServer @requestCallback

			resolve @listen()
		else
			workerLimit = os.cpus().length
			workerLimit = @cfg.server.workers if @cfg.server.workers < workerLimit

			# clean out old sockets
			if @cfg.server.method is 'socket'
				if fs.existsSync @cfg.server.socket
					fs.unlinkSync @cfg.server.socket

			@emit 'cluster.starting', this

			cluster.on 'online', => @emit.apply @lance, [ 'cluster.worker.online' ].concat toArray arguments
			cluster.on 'offline', => @emit.apply @lance, [ 'cluster.worker.exit' ].concat toArray arguments

			c = 0
			while c < workerLimit
				cluster.fork()
				++c

			resolve()

	listen: -> new Promise (resolve, reject) =>
		if @cfg.server.method is 'socket'
			@server.listen @cfg.server.socket, resolve
		else
			@server.listen @cfg.server.port, @cfg.server.host, resolve

	start: ->
		if @cfg.server.cluster
			return @createCluster()
		else
			@createServer()

			return @listen()

	relayEvents: (emitter) ->
		@helpers.events.relay this, emitter

	extendEmitter: (emittee) ->
		emittee.emitter	= @emitter
		merge emittee, emittee.emitter.__proto__
		emittee.on		= emittee.on.bind emittee.emitter
		emittee.emit	= emittee.emit.bind emittee.emitter

		return @emitter

	cfg		: require '../cfg'
	Emitter	: Emitter
	utils	: utils
	
	helpers	: require './helpers'

	Templater		: require './Templater'
	Router			: require './Router'
	Cache			: require './Cache'
	RequestHandler	: require './RequestHandler'

	merge this, this.prototype, 1
