path		= require 'path'
Promise		= require 'bluebird'
fs			= Promise.promisifyAll require 'fs'
cluster		= require 'cluster'
uglifyJs	= require 'uglify-js'
mkdirp		= Promise.promisify require 'mkdirp'
Browserify	= require 'browserify'
Coffeeify	= require 'coffeeify'
watch		= require 'watch'

{ clone, merge, typeOf, isAbsolute, changeExt, isExt, changeExt, exploreDir, empty, minify, minifyCss, escapeRegExp, toArray } = require './helpers/utils'
{ promisify } = require './helpers'

module.exports = Templater = ->
	self = this

	@constructor = (newCfg, @lance) =>
		oldCfg = require('../cfg').templater

		@cfg = clone.json oldCfg

		merge @cfg, newCfg if newCfg

		@cfg.root = @cfg.root or @lance.paths.root

		@locals = {}

		@cfg.bundles = @cfg.bundles or @cfg.bundle

		{ stylus, coffee, templater, css, js, assets, ect } = @cfg

		if css.inherit
			css.findIn or css.findIn = stylus.findIn
			css.saveTo or css.saveTo = stylus.saveTo

		if js.inherit
			js.findIn or js.findIn = coffee.findIn
			js.saveTo or js.saveTo = coffee.saveTo

		if stylus.inherit
			stylus.findIn or stylus.findIn = css.findIn
			stylus.saveTo or stylus.saveTo = css.saveTo

		if coffee.inherit
			coffee.findIn or coffee.findIn = js.findIn
			coffee.saveTo or coffee.saveTo = js.saveTo

		coffee.engine		= coffee.engine or require 'coffee-script'
		try stylus.engine	= stylus.engine or require 'stylus'

		try
			if not templater? and ect
				templater		= ect
				templater.name	= 'ect'

				if not templater.engine
					try templater.engine = require 'ect'

			if templater.name is 'ect'
				@templater.constructEct templater
		catch err
			throw new Error 'Invalid templating engine'

		css.match		= css.ext if not css.match
		js.match		= js.ext if not js.match
		stylus.match	= stylus.ext if not stylus.match
		coffee.match	= coffee.ext if not coffee.match

		models = { stylus, coffee, templater, css, js, assets }

		for key, model of models
			if match = model.match
				model.matchFn = if typeOf.String match
					(fileDir = '') -> return isExt fileDir, match
				else if typeOf.RegExp match
					(fileDir = '') -> return match.test fileDir
				else if typeOf.Function match
					match

			delete model.render if not typeOf.Function model.render
			delete model.watch if not typeOf.Function model.watch

		# Merge the configs straight over the top of the models, @stylus etc.
		for key, model of models
			merge this[ key ], model
	
		return this

	@initialize = =>
		await = [
			@bundle()
		]

		if @assets.findIn and @assets.saveTo
			await.push @assets.syncDirectory()

		return Promise.all await

	###
		Alias to `@templater.render`.
	###
	@render = (dir, locals) =>
		return @templater.render dir, locals

	###
		The templater, for templating files!
	###
	@templater =
		render: (dir = '', locals) ==>
			@lance.emit 'templater.render.template', dir, locals

			if locals
				locals = merge.black locals, clone.merge( @locals ), 6
			else
				locals = @locals

			rendered = yield new Promise (resolve, reject) =>
				@templater.engine.render dir, locals, (err, rendered = '') =>
					return reject err if err
					resolve rendered

			if @templater.minify and rendered
				rendered = minify rendered

			return rendered

		constructEct: (templater) =>
			if typeOf.Function templater.engine
				templater.options.root = if templater.findIn and typeOf.String templater.findIn
					@file.resolveToRoot templater.findIn
				else
					templater.options.root or @cfg.root

				templater.options.ext	= templater.options.ext or templater.ext
				templater.engine		= templater.engine templater.options

			templater.ext = templater.ext or templater.options.ext or '.ect'

	@stylus =
		render: (fileDir, destination) ==>
			fileDir = @file.resolve fileDir, @stylus.findIn

			@lance.emit 'templater.render.stylus', fileDir

			file = yield @file.read fileDir
		
			promise = new Promise (resolve, reject) =>
				try
					engine = @stylus.engine file, @stylus.options

					engine.set 'paths', [ path.dirname fileDir ]
					engine.set 'filename', fileDir

					dependencies = engine.deps()
				catch err then return reject err

				engine.render (err, rendered = '') =>
					return reject err if err
					return resolve rendered if not rendered

					newFileDir = @file.createSaveToPath destination or fileDir, {
						ext		: @css.ext
						findIn	: @stylus.findIn
						saveTo	: @stylus.saveTo
					}

					if dependencies
						@stylus.watchDependencies dependencies, fileDir, =>
							try
								@stylus.render fileDir, destination
							catch err
								console.error err

					if @stylus.minify
						rendered = minifyCss rendered

					resolve @file.write( newFileDir, rendered ).return engine

			promise.catch (err) =>
				@lance.emit 'error', err

		watchDependencies: (dependencies, fileDir, callback) =>
			listeners = @stylus.listeners[ fileDir ] or []

			dependencies.push fileDir

			for dep in dependencies when dep not in listeners
				listeners.push dep

				@watch.once dep
				@lance.on "templater.watch.change.#{dep}", callback

			return @stylus.listeners[ fileDir ] = listeners

		listeners: {}

	@coffee =
		render: (fileDir, destination) ==>
			fileDir = @file.resolve fileDir, @coffee.findIn

			@lance.emit 'templater.render.coffee', fileDir

			file = yield @file.read fileDir

			rendered = @coffee.engine.compile file

			newFileDir = @file.createSaveToPath destination or fileDir, {
				ext: @js.ext
				findIn: @coffee.findIn
				saveTo: @coffee.saveTo
			}

			if @coffee.minify
				try 
					if result = uglifyJs.minify rendered, { fromString: true, mangle: false }
						rendered = result.code

			@file.write newFileDir, rendered

	@js =
		render: (fileDir, destination) ==>
			fileDir = @file.resolve fileDir, @js.findIn

			@lance.emit 'templater.render.js', fileDir

			rendered = yield @file.read fileDir

			newFileDir = @file.createSaveToPath destination or fileDir, @js

			if @js.minify
				try
					if result = uglifyJs.minify rendered, { fromString: true, mangle: false }
						rendered = result.code

			@file.write newFileDir, rendered

	@css =
		render: (fileDir, destination) ==>
			fileDir = @file.resolve fileDir, @css.findIn

			@lance.emit 'templater.render.css', fileDir

			file = yield @file.read fileDir

			newFileDir = @file.createSaveToPath destination or fileDir, @css

			if @css.minify
				file = minifyCss file

			@file.write newFileDir, file

	@assets =
		watching: {}
		render: (fileDir, destination) ==>
			fileDir = @file.resolve fileDir, @assets.findIn

			@lance.emit 'templater.render.assets', fileDir

			if not ( yield @file.exists fileDir )
				throw new Error "#{fileDir} doesnt exist"

			newFileDir = @file.createSaveToPath destination or fileDir, @assets
			readStream = fs.createReadStream fileDir

			return @file.writeStream readStream, newFileDir

		syncDirectory: (root = @assets.findIn) ==>
			root = @file.resolveToRoot root

			await	= []
			o		= {}

			if @assets.matchFn?
				o.filter = (file) =>
					return @assets.matchFn file

			monitorDirectory = (filePath) =>
				return null if @assets.watching[ filePath ]
				@assets.watching[ filePath ] = true

				watch.createMonitor filePath, o, (monitor) =>
					for dir, file of monitor.files
						if file.isFile()
							await.push @assets.render dir

					monitor.on 'created', (fileDir, stats) =>
						@lance.emit 'templater.watch.created', fileDir, stats
						@lance.emit "templater.watch.created.#{fileDir}", fileDir, stats

						if stats.isFile() and @assets.matchFn fileDir
							@assets.render fileDir
						else if stats.isDirectory()
							@assets.syncDirectory fileDir

					monitor.on 'changed', (fileDir) =>
						@lance.emit 'templater.watch.change', fileDir
						@lance.emit "templater.watch.change.#{fileDir}", fileDir

						@assets.render fileDir

					monitor.on 'removed', (fileDir) =>
						@lance.emit 'templater.watch.removed', fileDir
						@lance.emit "templater.watch.removed.#{fileDir}", fileDir

			monitorDirectory root
			
			yield exploreDir root, {
				depth		: @cfg.depth or 8
				directory	: monitorDirectory
			}

			return yield Promise.all await

	@assets.find = (findIn, next) =>
		findIn = findIn or @assets.findIn
		@find findIn, @assets.match, next

	@assets.find.watch = (findIn) =>
		@assets.find findIn, (file, fileDir) =>
			@watch.asset fileDir

	@assets.find.render = (findIn) ==>
		await = []
		yield @assets.find findIn, (file, fileDir) =>
			await.push @assets.render fileDir

		return Promise.all await

	@assets.find.renderAndWatch = (findIn) ==>
		await = []
		yield @assets.find findIn, (file, fileDir) =>
			await.push @assets.render fileDir
			@assets.watch fileDir

		return Promise.all await

	#
	# Bundling
	#

	@bundle = (arg) =>
		arg = @cfg.bundles if not arg?

		await = []
		switch typeOf arg
			when 'object'
				for destination, files of arg
					await.push @bundle.render destination, files
			when 'array'
				for cfg in arg then do (cfg) =>
					destination	= cfg.destination or cfg.saveTo or cfg.to
					files		= cfg.files or cfg.file or cfg.source or cfg.from

					await.push @bundle.render destination, files
			else
				await.push @bundle.render.apply this, arguments

		return Promise.all await

	###
		Checks the first file in the files parameter to determine to render it toJs or toCss.
	###
	@bundle.render = (destination, files) ==>
		firstFile = if typeOf.Array files
			files[0]
		else
			files

		return if @file.isToJs firstFile
			yield @bundle.render.toJs.apply this, arguments
		else if @file.isToCss firstFile
			yield @bundle.render.toCss.apply this, arguments

	###
		Bundles up js/coffee-script with browserify.

		@param destination {String}
		@param files {String} {Array} Either one fileDir as a string or many in an array
	###
	@bundle.render.toJs = (destination, files) ==>
		originalArgs = arguments

		files = [ files ] if not typeOf.Array files

		for fileDir, index in files
			model			= @file.resolveModel fileDir
			files[index]	= @file.resolve fileDir, model.findIn

		b = Browserify files, @cfg.browserify

		b.on 'file', (fileDir) =>
			@bundle.render.toJs.watchDependencies [ fileDir ], destination, =>
				try
					@bundle.render.toJs destination, files
				catch err
					#console.log err
					@lance.emit 'error', err

		console.log 'js '.cyan, 'bundle'.grey, destination, files

		b.transform Coffeeify, { global: true }

		readStream = b.bundle().on 'error', (err) =>
			readStream.end()

		saveTo = @file.createSaveToPath destination

		return yield @file.writeStream readStream, saveTo

	@bundle.render.toJs.watchDependencies = (dependencies, fileDir, callback) =>
		listeners = @bundle.render.toJs.listeners[ fileDir ] or []

		for dep in dependencies when dep not in listeners
			listeners.push dep

			@watch.once dep
			@lance.on "templater.watch.change.#{dep}", callback

		return @bundle.render.toJs.listeners[ fileDir ] = listeners

	@bundle.render.toJs.listeners = {}

	###
		Bundles up css/stylus. Only accepts ONE file as `source`

		@param destination {String}
		@param source {String}
	###
	@bundle.render.toCss = (destination, source) ==>
		console.log 'css'.cyan, 'bundle'.grey, destination, source
		return yield @stylus.render source, destination

	#
	# Watch file operations
	#


	@watch = (fileDir, callback) =>
		return false if not fileDir

		@lance.emit 'templater.watch', fileDir

		fs.watch fileDir, (event) =>
			return false if event isnt 'change'

			if fileDir of @watch.timeouts
				clearTimeout @watch.timeouts[ fileDir ]

			@watch.timeouts[ fileDir ] = setTimeout =>
				@lance.emit 'templater.watch.change', fileDir, event
				@lance.emit "templater.watch.change.#{fileDir}", fileDir, event

				delete @watch.timeouts[ fileDir ]
				callback fileDir, event if callback
			, @cfg.watchTimeout


		return true

	@watch.timeouts = {}

	@watch.once = (fileDir, callback) =>
		if fileDir not of @watch.once.watching
			if @watch fileDir, callback
				@watch.once.watching[ fileDir ] = true

	@watch.once.watching = {}

	#
	# File operations
	#

	@file = {}

	@file.exists = (fileDir) => new Promise (resolve) -> fs.exists fileDir, resolve

	@file.deferredWrite = (key, fileDir, file) =>
		return new Promise (resolve) =>
			if cache = @file.deferredWrite.timeouts[ key ]
				cache.resolve()
			else
				cache = @file.deferredWrite.timeouts[ key ] = {}
			
			cache.fulfill = =>
				@file.write fileDir, file

			cache.resolve = resolve

			@file.deferredWrite.refresh key

	@file.deferredWrite.timeouts = {}

	@file.deferredWrite.refresh = (key, time = @cfg.writeTimeout) =>
		if cache = @file.deferredWrite.timeouts[ key ]
			clearTimeout cache.timeout if cache.timeout

			cache.timeout = setTimeout =>
				cache.resolve @file.deferredWrite.timeouts[ key ].fulfill()
				delete @file.deferredWrite.timeouts[ key ]
			, time

	@file.resolve = (dir = '', root = '') =>
		if not root
			if model = @file.resolveModel dir
				root = model.findIn
		
		root = @file.resolveToRoot root or @cfg.root

		return if isAbsolute dir then dir else path.join root, dir

	@file.resolveToRoot = (dir) =>
		return if isAbsolute dir then dir else path.join @cfg.root, dir

	@file.resolveModel = (fileDir) =>
		return switch
			when isExt fileDir, @coffee.ext then @coffee
			when isExt fileDir, @js.ext then @js
			when isExt fileDir, @stylus.ext then @stylus
			when isExt fileDir, @css.ext then @css
			when isExt fileDir, @templater.ext then @templater

	@file.read = (fileDir, encoding) ==>
		if not ( yield @file.exists fileDir )
			throw new Error "#{fileDir} doesnt exist"

		encoding = 'utf8' if not encoding and encoding isnt false
		return fs.readFileAsync fileDir, encoding

	@file.write = (fileDir, file) ==>

		yield @file.createDirectory fileDir

		result = yield fs.writeFileAsync fileDir, file
		@lance.emit 'templater.writeFile', fileDir

		return result

	@file.writeStream = (readStream, fileDir) ==>
		yield @file.createDirectory fileDir

		writeStream = fs.createWriteStream fileDir
		return new Promise (resolve, reject) =>
			errored = false
			error = (err) ->
				reject err if not errored
				errored = true

			readStream.once 'error', error
			writeStream.once 'error', error

			readStream.pipe writeStream
				.on 'error', error
				.on 'close', =>
					@lance.emit 'templater.writeFile', fileDir
					resolve()

	@file.createDirectory = (fileDir) ==>
		return mkdirp path.dirname( fileDir ), @cfg.writePermissions or ( 0o0755 & ( ~process.umask() ) )

	# By default, this preserve findIn rooted paths into the saveTo directory. Currently unchangable.
	@file.createSaveToPath = (fileDir, model) =>
		if not model
			model = @file.resolveModel fileDir

		{ findIn, saveTo, ext } = model

		findIn = findIn[0] if typeOf.Array findIn

		findIn	= @file.resolve findIn
		saveTo	= @file.resolve saveTo
		fileDir	= @file.resolve fileDir, findIn

		# todo: match the findIn array with the fileDir first?
		findIn = new RegExp ( '^' + escapeRegExp findIn )

		partial = if @cfg.preserveDirectory
			fileDir.replace findIn, ''
		else
			path.basename fileDir

		newFileDir = path.join saveTo, partial

		if ext
			newFileDir = changeExt newFileDir, ext

		return newFileDir

	@file.isToJs = (fileDir) => return isExt( fileDir, @coffee.ext ) or isExt( fileDir, @js.ext )
	@file.isToCss = (fileDir) => return isExt( fileDir, @stylus.ext ) or isExt( fileDir, @css.ext )

	promisify.all.generators this

	return @constructor.apply this, arguments

